// Copyright 2017-2025 WIT.COM Inc. All rights reserved.
// Use of this source code is governed by the GPL 3.0

package main

import (
	"fmt"
	"syscall"

	"github.com/awesome-gocui/gocui"
	"go.wit.com/log"
)

// register how the 'gocui' will work as a GO toolkit plugin
// all applications will use these keys. they are universal.

// tells 'gocui' where to send events
func registerHandlers(g *gocui.Gui) {
	defer func() {
		if r := recover(); r != nil {
			fmt.Fprintln(outf, "EVENT BINDINGS recovered in r", r)
			return
		}
	}()

	// mouse handlers
	g.SetKeybinding("", gocui.MouseLeft, gocui.ModNone, mouseDown)       // normal left mouse down
	g.SetKeybinding("", gocui.MouseLeft, gocui.ModMouseCtrl, ctrlDown)   // mouse with the ctrl key held down
	g.SetKeybinding("", gocui.MouseRelease, gocui.ModNone, mouseUp)      // mouse button release
	g.SetKeybinding("", gocui.MouseWheelUp, gocui.ModNone, wheelsUp)     // mouse button release
	g.SetKeybinding("", gocui.MouseWheelDown, gocui.ModNone, wheelsDown) // mouse button release

	// Ctrl key handlers
	g.SetKeybinding("", gocui.KeyCtrlC, gocui.ModNone, doExit)        // CTRL-C : exits the application
	g.SetKeybinding("", gocui.KeyCtrlV, gocui.ModNone, doPanic)       // CTRL-V : force a panic()
	g.SetKeybinding("", gocui.KeyCtrlD, gocui.ModNone, openDebuggger) // CTRL-D : open the (D)ebugger
	keyForced, modForced := gocui.MustParse("ctrl+z")                 // setup ctrl+z
	g.SetKeybinding("", keyForced, modForced, handle_ctrl_z)          // CTRL-Z :cleverly let's you background gocui (breaks cursor mouse on return)
	g.SetKeybinding("", gocui.KeyEsc, gocui.ModNone, doEsc)           // escape key

	// regular keys
	g.SetKeybinding("", 'H', gocui.ModNone, theHelp)                  // 'H' toggles on and off the help menu
	g.SetKeybinding("", 'O', gocui.ModNone, theStdout)                // 'O' toggle the STDOUT window
	g.SetKeybinding("", 'D', gocui.ModNone, theDarkness)              // 'D' toggles light/dark mode
	g.SetKeybinding("", 'q', gocui.ModNone, doExit)                   // 'q' exit
	g.SetKeybinding("", gocui.KeyTab, gocui.ModNone, tabCycleWindows) // '2' use this to test new ideas

	// stdout keys
	g.SetKeybinding("", gocui.KeyPgup, gocui.ModNone, stdoutPgup)           // Pgup scroll up the Stdout buffer
	g.SetKeybinding("", gocui.KeyPgdn, gocui.ModNone, stdoutPgdn)           // Pgdn scroll down the Stdout buffer
	g.SetKeybinding("", gocui.KeyHome, gocui.ModNone, stdoutHome)           // Pgdn scroll down the Stdout buffer
	g.SetKeybinding("", gocui.KeyArrowUp, gocui.ModNone, stdoutArrowUp)     // Pgdn scroll down the Stdout buffer
	g.SetKeybinding("", gocui.KeyArrowDown, gocui.ModNone, stdoutArrowDown) // Pgdn scroll down the Stdout buffer

	// debugging
	g.SetKeybinding("", '2', gocui.ModNone, theNotsure)            // '2' use this to test new ideas
	g.SetKeybinding("", 'S', gocui.ModNone, theSuperMouse)         // 'S' Super Mouse mode!
	g.SetKeybinding("", 'M', gocui.ModNone, printWidgetPlacements) // 'M' list all widgets with positions
	g.SetKeybinding("", 'L', gocui.ModNone, printWidgetTree)       // 'L' list all widgets in tree view
	g.SetKeybinding("", 'f', gocui.ModNone, theFind)               // 'f' shows what is under your mouse
	g.SetKeybinding("", 'd', gocui.ModNone, theLetterD)            // 'd' toggles on and off debugging buttons
	g.SetKeybinding("", 'q', gocui.ModNone, quit)                  // 'q' only exits gocui. plugin stays alive (?)
}

// flips on 'super mouse' mode  // this was awesome for debugging gocui. never remove this code.
// while this is turned on, it will print out every widget found under the mouse
func theSuperMouse(g *gocui.Gui, v *gocui.View) error {
	if me.supermouse {
		log.Log(GOCUI, "supermouse off")
		me.supermouse = false
	} else {
		me.supermouse = true
		log.Log(GOCUI, "supermouse on")
	}
	return nil
}

// use this to test code ideas // put whatever you want here and hit '2' to activate it
func theNotsure(g *gocui.Gui, v *gocui.View) error {
	log.Info("got to theNotsure(). now what? dark =", me.dark)
	me.refresh = true
	/*
		if me.debug {
			log.Info("debugging off")
			me.debug = false
		} else {
			log.Info("debugging on")
			me.debug = true
		}
		win := findWindowUnderMouse()
		if win != nil {
			win.dumpWidget("found() win. redrawing window:")
			win.makeWindowActive()
		}
	*/
	return nil
}

func theDarkness(g *gocui.Gui, v *gocui.View) error {
	if me.dark {
		me.dark = false
		log.Info("you have seen the light")
	} else {
		me.dark = true
		log.Info("you have entered into darkness")
	}
	return nil
}

func wheelsUp(g *gocui.Gui, v *gocui.View) error {
	// log.Info("private wheels up")
	me.stdout.pager -= 2
	if me.stdout.pager < 0 {
		me.stdout.pager = 0
	}
	me.stdout.tk.refreshStdout()
	return nil
}

func wheelsDown(g *gocui.Gui, v *gocui.View) error {
	// log.Info("you've landed")
	me.stdout.pager += 2
	if me.stdout.pager > len(me.stdout.outputS) {
		me.stdout.pager = len(me.stdout.outputS)
	}
	me.stdout.tk.refreshStdout()
	return nil
}

func tabCycleWindows(g *gocui.Gui, v *gocui.View) error {
	// log.Info("try to switch windows here")
	if len(me.allwin) != len(findWindows()) {
		me.allwin = findWindows()
	}
	tk := findNextWindow()
	if tk == nil {
		log.Info("findNextWindow() err. returned nil")
		return nil
	}
	tk.makeWindowActive()
	return nil
}

func doEsc(g *gocui.Gui, v *gocui.View) error {
	log.Info("got escape key")
	if me.dropdown.active {
		me.dropdown.tk.Hide()
		me.dropdown.active = false
		log.Info("escaped from dropdown")
	}
	if me.textbox.active {
		me.textbox.tk.Hide()
		me.textbox.active = false
		log.Info("escaped from textbox")
	}
	return nil
}

func theShow(g *gocui.Gui, v *gocui.View) error {
	var w *guiWidget
	w = me.treeRoot.TK.(*guiWidget)
	w.showWidgets()
	return nil
}

func doExit(g *gocui.Gui, v *gocui.View) error {
	standardExit()
	return nil
}

func doPanic(g *gocui.Gui, v *gocui.View) error {
	log.Log(GOCUI, "do panic() here")
	standardClose()
	panic("forced panic in gocui")
}

func openDebuggger(g *gocui.Gui, v *gocui.View) error {
	me.myTree.SendEnableDebugger()
	return nil
}

func theFind(g *gocui.Gui, v *gocui.View) error {
	w, h := g.MousePosition()
	for _, tk := range findByXY(w, h) {
		// tk.v.BgColor = gocui.ColorGreen
		tk.dumpWidget("theFind()")
		// tk.verifyRect()
	}
	return nil
}

// is run whenever anyone hits 'd' (in an open space)
func theLetterD(g *gocui.Gui, v *gocui.View) error {
	// widgets that don't have physical existance in
	// a display toolkit are hidden. In the case
	// of gocui, they are set as not 'visible' and put offscreen
	// or have the size set to zero
	// (hopefully anyway) lots of things with the toolkit
	// still don't work

	fakeStartWidth = me.FakeW
	fakeStartHeight = me.TabH + me.FramePadH
	if me.showDebug {
		showFake()
		me.showDebug = false
	} else {
		hideFake()
		me.showDebug = true
	}
	return nil
}

func theHelp(g *gocui.Gui, v *gocui.View) error {
	if me.showHelp {
		log.Info("Show the help!")
		showHelp()
	} else {
		log.Info("Hide the help!")
		hideHelp()
	}
	return nil
}

// todo: find and give credit to the person that I found this patch in their forked repo
// handle ctrl+z
func handle_ctrl_z(g *gocui.Gui, v *gocui.View) error {
	gocui.Suspend()
	log.Info("got ctrl+z")
	syscall.Kill(syscall.Getpid(), syscall.SIGSTOP)
	log.Info("got ctrl+z syscall() done")
	gocui.Resume()
	return nil
}
